package src.BAALL

import java.net.Socket
import java.util
import scala.collection.immutable.Queue
import scala.io.Source
import scala.xml.pull.XMLEventReader
import scala.xml.{Node, Elem, XML}
import src.ship.JavaLogging

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 8/1/12
 * Time: 2:06 PM
 * This class defines utility function to communicate via linknx with an environment, e.g., the BAALL
 */

/**
 * internal representation for objects including there state. at the moment, the baall uses only the following types
 * 1.001: switching (on/off) (EIS1)
 * 5.xxx: 8bit unsigned integer (from 0 to 255) (EIS6)
 * 16.000: string (max 14 ASCII chars restricted to ASCII codes 0..127) (EIS15)
 */

sealed trait LKNXObject {
  def name: String
}

case class SwitchObject(name: String, value: Boolean) extends LKNXObject

case class ValObject(name: String, value: Int) extends LKNXObject

case class StringObject(name: String, value: String) extends LKNXObject

case class PosObject(name: String, pos: String) extends LKNXObject

object LKNXType extends Enumeration {
  type LKNXType = Value
  val TSwitchObject, TStringObject, TValObject, TPosObject = Value
}

object PosPrefix {
  val posPrefix: String = "pos_"

  def strip(pos: String): String =
    if (pos.length > posPrefix.length && pos.startsWith(posPrefix) &&
      pos.charAt(posPrefix.length).isDigit)
      pos.substring(posPrefix.length)
    else pos

  def add(pos: String): String = if (!pos.isEmpty && pos.head.isDigit) posPrefix + pos else pos
}

object EOT {
  val byte: Byte = 4
}

object RouteGraph {
  var xml: Elem = XML.load(this.getClass.getClassLoader.getResource("RouteGraph.xml"))
}

class BaallSocket(host: String, port: Int) extends JavaLogging {
  var skt: Option[Socket] = None
  var rdr: Option[XMLEventReader] = None

  def connect(descr: String) {
    if (skt.isEmpty) {
      try {
        val sk = new Socket(host, port)
        rdr = Some(new XMLEventReader(Source.fromInputStream(sk.getInputStream)))
        skt = Some(sk)
      } catch {
        case e: Exception => log.warning(descr + " connect exception " + host + ":" + port + " " + e)
      }
    }
  }

  def sendStringAsBytes(msg: String) {
    skt match {
      case Some(sk) =>
        val out = sk.getOutputStream
        out.write(msg.getBytes)
        out.write(EOT.byte)
        out.flush()
      case None => log.warning("not connected for sending " + host + ":" + port)
    }
  }

  def readXml(): Seq[Node] = {
    if (skt.isDefined) XmlRead.getXML(XmlRead.getNextEvents(host + ":" + port, rdr.get))
    else {
      log.warning("not connected for read " + host + ":" + port)
      Nil
    }
  }
}

/**
 * represents a connection to a Linknx server; provides functionality to retrieve information about the objects and
 * to update their values
 * @param server name/IP of the LKNX server
 * @param port port of the LKNX server
 */
class Linknx(server: String, port: Int) extends JavaLogging {

  import LKNXType._

  val skt: BaallSocket = new BaallSocket(server, port)

  /**
   * Stores the counter value of the last read button event
   */

  var lastReadButtonEvent = 0

  /**
   * establishes the connection with the server
   */
  def connect() {
    skt.connect("Linknx")
  }

  /**
   * disconnects from the server
   */
  def disconnect(orig: String, e: Exception) {
    log.warning(orig + "connect Exception on " + server + ":" + port + ": " + e)
    skt.skt = None
  }

  /**
   * read but do not process the result
   */
  def skipresult() {
    log.info("Wait for response...")
    try {
      val res = skt.readXml()
      log.info("response done " + res)
    }
    catch {
      case e: Exception => disconnect("skipresult ", e)
    }
  }

  /**
   * returns the status of all registered objects in the environment
   */
  def readAllObjects(): Seq[LKNXObject] = {
    skt.skt match {
      case Some(_) => this.synchronized {
        try {
          val m = "<read><objects/></read>"
          //log.info(m)
          skt.sendStringAsBytes(m)
          val x = skt.readXml()
          x flatMap processRead
        } catch {
          case e: Exception =>
            disconnect("readAllObjects ", e)
            Nil
        }
      }
      case None => {
        connect()
        if (skt.skt.isDefined) readAllObjects() else Nil
      }
    }
  }

  /**
   * reads last button press
   */

  def readLastButton(): Seq[BAALLButtonEvent] = {
    skt.skt match {
      case Some(_) => this.synchronized {
        try {
          val m = "<read><button/></read>"

          skt.sendStringAsBytes(m)
          val x = skt.readXml()
          log.info("READ " + x)
          x flatMap processButton
        } catch {
          case e: Exception =>
            disconnect("readLastButton ", e)
            Nil
        }
      }
      case None => {
        connect()
        if (skt.skt.isDefined) readLastButton() else Nil
      }
    }
  }

  /**
   * processes a read xml return value
   * @param input linknx xml representation of a read result
   */
  def processRead(input: xml.Node): Seq[LKNXObject] = {
    //log.warning("From BAALLServer/Yamamoto Processing "+input)
    input match {
      case <read>{contents@_*}</read> => contents flatMap {
          case <objects>{ccontents@_*}</objects> => ccontents flatMap processObject
          case _ => Nil
        }
      case _ => Nil
    }
  }

  /**
   * processes an objects specification
   * @param input linknx xml representation of the object to be processed
   */
  def processObject(input: xml.Node): Seq[LKNXObject] = {
    input match {
      case <object>{contents@_*}</object> => Seq[LKNXObject](parseObject(input))
      case _ => Nil
    }
  }

  /**
   * parses an object and adds it to the hashmap
   * @param input the new object in
   * @return the parsed LKNXObject
   */
  def parseObject(input: Node): LKNXObject = {
    //log.info("parseObject "+input)
    val value: String = input.attribute("value").get.text
    val id: String = input.attribute("id").get.text
    val objecttype = guessType(id, value)
    objecttype match {
      case TSwitchObject => SwitchObject(id, value == "on")
      case TValObject => ValObject(id, value.toInt)
      case TStringObject => StringObject(id, value)
      case TPosObject => {
        val newvalue = PosPrefix.add(value)
        PosObject(id, newvalue)
      }
    }
  }

  /**
   * reads the configuration
   */
  def readConfig() {
    this.synchronized {
      skt.skt match {
        case Some(_) =>
          skt.sendStringAsBytes("<read><config/></read>")
          skt.readXml() map processRead
        case None => connect(); readConfig()
      }
    }
  }

  /**
   * processes an button prressed information
   * @param input baall-server/linknx xml representation of the button pressed information to be processed
   */
  def processButton(input: xml.Node): Seq[BAALLButtonEvent] = {
    input match {
      case <read>{_*}</read> =>
        val buttons = input \ "button"
        buttons flatMap {
          case button@(<button>{_*}</button>) =>
            //println("Button "+button)
            val buttoncounter = (button \ "count").text.toInt
            if (buttoncounter > lastReadButtonEvent) {
              lastReadButtonEvent = buttoncounter
              val buttonid = (button \ "id").text
              val buttontarget = (button \ "dest").text
              val buttonvalue = (button \ "value").text
              val buttontime = util.Calendar.getInstance().getTimeInMillis
              Seq[BAALLButtonEvent](BAALLButtonEvent(buttonid, buttontarget, buttonvalue, buttontime))
            } else Nil
          case _ => Nil
        }
      case _ => Nil
    }
  }

  var writeQueue = Queue[String]()

  /**
   * sends a status update via the protocol
   * @param id name of the object
   * @param value value as string
   */
  def writeObject(id: String, value: String) {
    this.synchronized {
      /* println("<write><object id=\"" + id + "\" value=\"" + value + "\"/></write>")
      skt.getOutputStream.write(("<write><object id=\"" + id + "\" value=\"" + value + "\"/></write>").getBytes);
      skt.getOutputStream.write(EOT.byte)
      skt.getOutputStream.flush()
      skipresult() */
      writeQueue = writeQueue.enqueue("<object id=\"" + id + "\" value=\"" + PosPrefix.strip(value) + "\"/>")
    }
  }

  /**
   * sends a status update via the protocol
   * @param id name of the object
   * @param p value as Position
   */
  def write2DPosition(id: String, p: Position) {
    val Position(x, y, r) = p
    this.synchronized {
      /* println("<write><object id=\"" + id + "\" value=\"" + value + "\"/></write>")
      skt.getOutputStream.write(("<write><object id=\"" + id + "\" value=\"" + value + "\"/></write>").getBytes);
      skt.getOutputStream.write(EOT.byte)
      skt.getOutputStream.flush()
      skipresult() */
      val msg = "<position2D id=\"" + id + "\" x=\"" + x + "\" y=\"" + y + "\" rotation=\"" + r + "\"/>"
      log.info("Sending out " + msg)
      writeQueue = writeQueue.enqueue(msg)
    }
  }

  def sendRouteGraph() {
    val string = RouteGraph.xml.toString()
    // log.info("route graph: " + string)
    skt.sendStringAsBytes(string)
  }

  def sendWriteQueue() {
    this.synchronized {
      skt.skt match {
        case Some(_) =>
          if (!writeQueue.isEmpty) {
            var string = "<write>"
            while (!writeQueue.isEmpty) {
              val (s, r) = writeQueue.dequeue
              string = string + s
              writeQueue = r
            }
            string = string + "</write>"
            log.info(string)
            skt.sendStringAsBytes(string)
            skipresult()
          }
        case None => {
          connect()
          if (skt.skt.isDefined) sendWriteQueue()
        }
      }
    }
  }

  /**
   * guesses the type based on its value
   * @param value text contained in the value attribute of the object
   * @return the guessed type of the object based on the value
   */
  def guessType(id: String, value: String): LKNXType = {
    //println("id "+id+" "+id.toLowerCase()+" "+id.toLowerCase.startsWith("rolland"))
    if (id.toLowerCase.startsWith("rolland")) TPosObject
    else if (id.toLowerCase.startsWith("paul")) TPosObject
    else if (id.toLowerCase.startsWith("eve")) TPosObject
    else if (value == "off" | value == "on") TSwitchObject
    else if (value.length > 0 && value.charAt(0).isLetter) TStringObject
    else {
      try {
        value.toInt
        TValObject
      }
      catch {
        case e: Throwable => log.warning("no integer value for id %s = '%s' : %s".format(id,value,e))
          TStringObject
      }
    }
  }

  /**
   * returns the name of a lknx-object
   * @param obj lknx-objec
   * @return
   */
  def getName(obj: LKNXObject): String = {
    obj match {
      case PosObject(name, pos) => name
      case ValObject(name, value) => name
      case StringObject(name, value) => name
      case SwitchObject(name, value) => name
    }
  }

  /**
   * returns the name of a lknx-object
   * @param obj lknx-object
   * @return
   */
  def getVal(obj: LKNXObject): String = {
    obj match {
      case PosObject(name, pos) => pos
      case ValObject(name, value) => value.toString
      case StringObject(name, value) => value
      case SwitchObject(name, value) => if (value) "on" else "off"
    }
  }
}
