package src.BAALL

import akka.actor.ActorDSL._
import akka.actor.{ActorLogging, ReceiveTimeout, ActorRef}
import scala.List
import scala.Seq
import scala.Some
import scala.concurrent.duration._
import scala.xml.{Elem, Node}
import src.math.{Point, Segment}
import src.ship._
import src.ship.external.PathCoordination

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 23.01.2013
 * Time: 14:14
 */

// Test in Scala Console: var ro = src.BAALL.RollandObjectDecoder.testT2

object RollandObjectDecoder1 {

  def fromXML(cmd: Node): List[RollandObject] = {
    def attrMatch(d: Node, attrname: String): Option[String] = {
      if ((d \ ("@" + attrname)).isEmpty) None
      else Some((d \ ("@" + attrname)).toString())
    }

    cmd match {
      case d@ <OdometryPose2D/> =>
        (attrMatch(d, "x"), attrMatch(d, "y"), attrMatch(d, "rotation")) match {
          case (Some(x), Some(y), Some(rotation))
          => List(OdometryPose2D(x, y, rotation))
          case _ => Nil
        }
      case d@ <LocalizedPose2D/> =>
        (attrMatch(d, "x"), attrMatch(d, "y"), attrMatch(d, "rotation")) match {
          case (Some(x), Some(y), Some(rotation))
          => List(LocalizedPose2D(x, y, rotation))
          case _ => Nil
        }
      case d@ <TargetReachedState/> =>
        attrMatch(d, "trs") match {
          case Some(s) => List(TargetReachedState(trs = s == "TRUE"))
          case None => List.empty
        }
      case d@ <DrivingDestination>{ds@_*}</DrivingDestination> =>
        val nodelist = ds.toList
        val cmdlist = nodelist flatMap ((n: Node) => fromXML(n).filter(_.isInstanceOf[NavigationGraphEdge]).asInstanceOf[List[NavigationGraphEdge]])

        (attrMatch(d, "label"), attrMatch(d, "x"), attrMatch(d, "y")) match {
          case (Some(l), Some(x), Some(y)) => List(DrivingDestination(PosPrefix.add(l), x, y, cmdlist.sortWith((e1, e2) => e1.targetNode <= e2.targetNode)))
          case _ => List.empty
        }
      case d@ <NavigationGraphEdge/> =>
        (attrMatch(d, "direction"), attrMatch(d, "type"), attrMatch(d, "targetNode")) match {
          case (Some(dir), Some(tp), Some(tn)) => List(NavigationGraphEdge(dir, tp, PosPrefix.add(tn)))
          case _ => List.empty
        }
      case <DrivingDestinations>{ds@_*}</DrivingDestinations> =>
        val nodelist = ds.toList
        val cmdlist = nodelist flatMap ((n: Node) => fromXML(n).filter(_.isInstanceOf[DrivingDestination]).asInstanceOf[List[DrivingDestination]])
        List(DrivingDestinations(cmdlist.sortWith((d1, d2) => d1.label <= d2.label)))
      case d@ <SetNavigationDestination/> =>
        attrMatch(d, "label") match {
          case Some(l) => List(SetNavigationDestination(PosPrefix.add(l)))
          case _ => List.empty
        }
      case d@ <NavigationDestination/> =>
        attrMatch(d, "label") match {
          case Some(l) => List(NavigationDestination(PosPrefix.add(l)))
          case _ => List.empty
        }
      case d@ <FRAMEINFO/> =>
        (attrMatch(d, "Version"), attrMatch(d, "Time")) match {
          case (Some(v), Some(t)) => List(FrameInfo(v, t))
          case _ => List.empty
        }
      case d@ <SetRemoteDrive/> =>
        (attrMatch(d, "speed"), attrMatch(d, "steering")) match {
          case (Some(sp), Some(st)) => List(SetRemoteDrive(sp, st))
          case _ => List.empty
        }
      case d@ <POSITION/> =>
        (attrMatch(d, "Rotation"), attrMatch(d, "x"), attrMatch(d, "y")) match {
          case (Some(r), Some(x), Some(y)) => List(Position(x, y, r))
          case _ => List.empty
        }
      case s: scala.xml.Text => List()
      case _ => List()
    }
  }
}

class ConflictMap extends JavaLogging {

  override def toString: String = {
    val nodeedgeconflicts =
      (nodeEdgeConflictMap.iterator :\ 0)((entry, sum) => sum + entry._2.size)
    val nodenodeconflicts =
      (nodeNodeConflictMap.iterator :\ 0)((entry, sum) => sum + entry._2.size)
    "ConflictMap(Node/Edge=" + nodeedgeconflicts + ", Node/Node=" + nodenodeconflicts + ")"
  }

  var nodeEdgeConflictMap: Map[String, Set[(String, String)]] = Map.empty
  var nodeNodeConflictMap: Map[String, Set[String]] = Map.empty

  def getEdgeConflicts(d: String): Set[(String, String)] =
    nodeEdgeConflictMap.getOrElse(d, Set.empty)

  def getNodeConflicts(d: String): Set[String] =
    nodeNodeConflictMap.getOrElse(d, Set.empty)

  def nodeEdgeConflictUpdated(d: String, e: (String, String)): ConflictMap = {
    val x = nodeEdgeConflictMap updated(d, getEdgeConflicts(d) + e)
    val y = nodeNodeConflictMap
    new ConflictMap {
      nodeEdgeConflictMap = x
      nodeNodeConflictMap = y
    }
  }

  def edgeConflictsDeleted: ConflictMap = {
    var x = nodeEdgeConflictMap
    val y = nodeNodeConflictMap
    NonConflicts.nodeEdgeRaw.map {
      case (n, (s, t)) =>
        val k = PosPrefix.add(n)
        x.get(k) match {
          case Some(l) =>
            val p = (PosPrefix.add(s), PosPrefix.add(t))
            if (l.contains(p))
              x = x.updated(k, l - p)
          // else log.debug("no conflict edge " + p + " with node " + k)
          case _ => log.info("no conflict with " + k)
        }
    }
    new ConflictMap {
      nodeEdgeConflictMap = x
      nodeNodeConflictMap = y
    }
  }

  def nodeNodeConflictUpdated(d1: String, d2: String): ConflictMap = {
    val x = nodeEdgeConflictMap
    val y = nodeNodeConflictMap updated(d1, getNodeConflicts(d1) + d2)
    new ConflictMap {
      nodeEdgeConflictMap = x
      nodeNodeConflictMap = y
    }
  }

  def conflicts(p: String, start: String, end: String): Boolean = {
    nodeEdgeConflictMap.get(p) match {
      case None => false
      case Some(l) =>
        l.exists(entry => entry._1 == start && entry._2 == end || entry._1 == end && entry._2 == start)
    }
  }

  def conflicts(n1: String, n2: String): Boolean = {
    nodeNodeConflictMap.get(n1) match {
      case None => false
      case Some(dlist) => dlist.exists(_ == n2)
    }
  }
}

sealed abstract class RollandObject {
  def toXML: Elem
}

// <TargetReachedState trs="TRUE|FALSE">

case class OdometryPose2D(x: String, y: String, rotation: String) extends RollandObject {
  override def toXML: Elem = <OdometryPose2D x={x} y={y} rotation={rotation}/>
}

case class LocalizedPose2D(x: String, y: String, rotation: String) extends RollandObject {
  override def toXML: Elem = <LocalizedPose2D x={x} y={y} rotation={rotation}/>
}

case class TargetReachedState(trs: Boolean) extends RollandObject {
  override def toXML: Elem =
      <TargetReachedState trs={if (trs) "TRUE" else "FALSE"}/>
}

case class DrivingDestination(label: String, x: String, y: String, edges: List[NavigationGraphEdge]) extends RollandObject {
  override def toXML: Elem =
    <DrivingDestination label={PosPrefix.strip(label)} x={x} y={y}>
      {edges map (_.toXML)}
    </DrivingDestination>
}

case class NavigationGraphEdge(direction: String, edgetype: String, targetNode: String) extends RollandObject {
  override def toXML: Elem =
      <NavigationGraphEdge direction={direction} type={edgetype} targetNode={PosPrefix.strip(targetNode)}/>
}

case class DrivingDestinations(destinations: List[DrivingDestination]) extends RollandObject with JavaLogging {
  override def toXML: Elem =
    <DrivingDestinations>
      {destinations map (_.toXML)}
    </DrivingDestinations>

  def findClosestDrivingDestination(pos: Position): Option[DrivingDestination] = {
    val parse2Double = (s: String) => try {
      s.toDouble
    } catch {
      case e: Exception => log.warning("Error transforming to double " + s + ": " + e); 0.0
    }
    val pospoint = Point(parse2Double(pos.x), parse2Double(pos.y))
    val distanceToCurrent = (a: DrivingDestination) => {
      Point(parse2Double(a.x), parse2Double(a.y)).distanceTo(pospoint)
    }

    (destinations :\ (None: Option[DrivingDestination]))((d, current) => current match {
      case None => Some(d)
      case Some(c) => if (distanceToCurrent(c) < distanceToCurrent(d)) Some(c) else Some(d)
    })
  }

  def findNode(name: String): Option[DrivingDestination] =
    destinations find ((n) => n.label == name)

  def conflicts: ConflictMap = {
    conflictingNodeEdges(1000).edgeConflictsDeleted
  }

  def conflictingNodeEdges(minDist: Double): ConflictMap = {
    var conflicts: ConflictMap = new ConflictMap
    def toPoint(d: DrivingDestination): Point = Point(d.x.toDouble, d.y.toDouble)
    for (a <- destinations) {
      val pointA = toPoint(a)
      val aLabel = a.label
      for (b <- destinations.filter(_ != a)) {
        val pointB = toPoint(b)
        val bLabel = b.label
        for (e <- a.edges) {
          val cLabel = e.targetNode
          findNode(cLabel) match {
            case Some(c) => if (cLabel != bLabel) {
              val pointC = toPoint(c)
              val d = Segment(pointA, pointC).distanceTo(pointB)
              if (d <= minDist) {
                // log.debug("Node " + bLabel + " is only " + d + " from segment " + aLabel + " to " + cLabel)
                conflicts = conflicts.nodeEdgeConflictUpdated(bLabel, (aLabel, cLabel))
              }
            }
            case _ => log.warning("point " + cLabel + " unknown")
          }
        }
        val d = pointB.distanceTo(pointA)
        if (d <= minDist) {
          // log.debug("Node " + aLabel + " is only " + d + " from node " + bLabel)
          conflicts = conflicts.nodeNodeConflictUpdated(aLabel, bLabel)
          conflicts = conflicts.nodeNodeConflictUpdated(bLabel, aLabel)
        }
      }
    }
    conflicts
  }
}

case class SetNavigationDestination(label: String) extends RollandObject with RollandCMD {
  override def toXML: Elem = <SetNavigationDestination label={PosPrefix.strip(label)}/>
}

case class NavigationDestination(label: String) extends RollandObject {
  override def toXML: Elem = <NavigationDestination label={PosPrefix.strip(label)}/>
}

case class FrameInfo(version: String, time: String) extends RollandObject {
  override def toXML: Elem = <FRAMEINFO Version={version} Time={time}/>
}

case class SetRemoteDrive(speed: String, steering: String) extends RollandObject {
  override def toXML: Elem = <SetRemoteDrive speed={speed} steering={steering}/>
}

case class Position(x: String, y: String, rotation: String) extends RollandObject {
  override def toXML: Elem = <POSITION Rotation={rotation} x={x} y={y}/>
}

trait RollandCMD

case class VirtualID(s: String) extends RollandCMD

class RollandConfiguration {
  var drivingDestinations: List[DrivingDestination] = {
    val node = RollandObjectDecoder1.fromXML(RouteGraph.xml)
    node.last.asInstanceOf[DrivingDestinations].destinations
  }

  var currentNavigationDestination: Option[String] = None
  var position2D: Option[Position] = None
  var position: Option[String] = None
  var virtualid: Option[String] = None

  def reset() {
    currentNavigationDestination = None
    virtualid = None
  }
}

case class RollandClient(host: String, port: Int, name: String, actors: Set[ActorRef])
  extends ActWithStash with ActorLogging {
  val configuration = new RollandConfiguration {
    virtualid = Some(name)
  }
  val socket: BaallSocket = new BaallSocket(host, port)
  var listeners: Set[ActorRef] = actors

  def connect() {
    log.info("Rolland connect request to " + host + ":" + port)
    socket.connect("Rolland")
  }

  private def resetSocket() {
    socket.skt match {
      case Some(s) => try {
        s.close()
      } catch {
        case e: Exception => log.warning(host + ":" + port + " reset exception " + e)
      }
      case _ =>
    }
    socket.skt = None
    configuration.position2D = None
  }

  def registerListener(a: ActorRef) {
    listeners = listeners + a
  }

  def unregisterListener(a: ActorRef) {
    listeners = listeners - a
  }

  def setvirtualname(s: String) {
    configuration.virtualid = Some(s)
    informListenersAboutCurrentPosition()
  }

  private def send(s: Seq[Node]) {
    if (socket.skt.isDefined) {
      try {
        sendRaw(s.toString())
      } catch {
        case e: Exception => log.warning("send connection reset " + e); resetSocket()
      }
    } else {
      connect()
      send(s)
    }
  }

  private def sendRaw(s: String) {
    val m = "<?xml version=\"1.0\" encoding=\"" + xml.XML.encoding + "\"?>" + s
    log.info("Send2Rolland " + configuration.virtualid + " " + m)
    socket.sendStringAsBytes(m)
  }

  private def sendURC(cmds: List[RollandObject]) {
    val rollandmsg = <URC>{cmds map (_.toXML)}</URC>
    send(rollandmsg)
    //println("Sent to Rolland "+configuration.virtualid+": \n"+rollandmsg)
  }

  private def receiveFromRolland(): Option[Seq[Node]] = {
    if (socket.skt.isEmpty) {
      connect()
      receiveFromRolland()
    } else {
      try {
        socket.sendStringAsBytes("")
        if (socket.rdr.get.available()) {
          //XmlRead.hasMore(configuration.virtualid,fromRolland.get)) {
          log.info(configuration.virtualid + " before read ")
          val x = socket.readXml()
          log.info(configuration.virtualid + " sent " + x)
          x match {
            case _ :: _ => Some(x)
            case _ => None
          }
        } else {
          None
        }
      } catch {
        case e: Exception => log.warning("receive connection reset " + e); resetSocket(); None
      }
    }
  }

  def parseURC(s: Node): List[RollandObject] = {
    s match {
      case <URC>{cmds@_*}</URC> => {
        cmds.toList flatMap ((n: Node) => RollandObjectDecoder1.fromXML(n))
      }
      case _ => log.info("Ignoring Rolland Message " + s); List.empty
    }
  }

  val atRole = NamedRole("at")
  val nextposRole = NamedRole("wcnextposition")
  val posstatusRole = NamedRole("position_status")
  val unassignedConcept = ConceptConstant("Unassigned")
  var waiting4TargetReachedStateChange: Boolean = false
  var lastTargetReachedState: Option[Boolean] = None
  var lastInformed2DPosition: Long = 0

  def obtainRollandInfo() {
    receiveFromRolland() match {
      case Some(xml) => {
        //println("Received from Rolland \n")
        //xml map (x => println(ppxml.format(x)))
        xml map {y => parseURC(y) map {
            //RollandObjectDecoder.fromXML(x) match {
            case LocalizedPose2D(xpos, ypos, rotation) =>
              val p = Position(xpos, ypos, rotation)
              log.info(configuration.virtualid + " received position: " + p)
              (configuration.position2D, configuration.currentNavigationDestination) match {
                case (None, Some(s)) =>
                  sendURC(List(SetNavigationDestination(s)))
                  waiting4TargetReachedStateChange = true
                  log.info("After obtaining position2d send SetNavigation: " + s)
                case _ =>
                  log.info("After obtaining position2d, nothing to send to Rolland ")
              }
              configuration.position2D = Some(p)
              (configuration.virtualid, configuration.position2D) match {
                case (Some(n), Some(pos)) =>
                  if ((System.currentTimeMillis() - lastInformed2DPosition) >= 1000) {
                    log.info("Sending new 2D position " + pos)
                    SHIPTool.ref ! RollandPosition2D(n, pos)
                    lastInformed2DPosition = System.currentTimeMillis()
                  }
                case _ =>
              }

            /*println(configuration.virtualid+": "+configuration.position2D+" closest position "+
              DrivingDestinations(configuration.drivingDestinations).findClosestDrivingDestination(p))  */
            case DrivingDestinations(nodes) =>
              configuration.drivingDestinations = nodes
              log.info(configuration.virtualid + " received new route graph: " + nodes)
              //PathCoordination.setCurrentConflictMap(DrivingDestinations(nodes).conflictingNodeEdges(750))
              PathCoordination.setCurrentConflictMap(DrivingDestinations(nodes).conflicts)
            /*(configuration.virtualid,configuration.drivingDestinations) match {
              case (Some(n),destlist) => SHIPTool ! RollandNewRouteGraph(n,DrivingDestinations(destlist))
              case _ =>
            } */

            // TODO: Update the graph in the ontology as well and send it as an update to the listeners
            //case NavigationDestination(label)::_ =>
            case (v@TargetReachedState(trs)) =>
              log.info(configuration.virtualid + " received " + v + "\nwaiting4target = " + waiting4TargetReachedStateChange)

              if (waiting4TargetReachedStateChange && Some(trs) != lastTargetReachedState) {
                waiting4TargetReachedStateChange = false
              }
              lastTargetReachedState = Some(trs)
              //println("Got "+v)
              if (trs && !waiting4TargetReachedStateChange) {
                //println("configuration.currentNavigationDestination "+configuration.currentNavigationDestination)
                //println("configuration.position = "+configuration.position)
                configuration.currentNavigationDestination match {
                  case Some(label) =>
                    configuration.position match {
                      case Some(l) if l != label =>
                        configuration.position = Some(label)
                        informListenersAboutCurrentPosition()
                      case None =>
                        configuration.position = Some(label)
                        informListenersAboutCurrentPosition(withstatus = true)
                      case _ =>
                    }
                  //configuration.currentNavigationDestination = None
                  case None => // there was no target, so let's try to determine current position
                    // by closest drivingdestination
                    configuration.position2D match {
                      case Some(p) =>
                        DrivingDestinations(configuration.drivingDestinations).findClosestDrivingDestination(p) match {
                          case Some(d) => if (Some(d.label) != configuration.position) {
                            log.info(configuration.virtualid + " found " + d + " to be closest to " + p)
                            configuration.position = Some(d.label)
                            informListenersAboutCurrentPosition(withstatus = true)
                          }
                          case None => log.info("no closest point found " + p)
                        }
                      case None => log.info("no current position known")
                    }
                }
              }
            case x => log.info("received from rolland: " + x)
          }
        }
        //xml map ((x:Node) => for(listener <- listeners) listener ! x)
      }
      case None =>
    }
  }

  def setNaviDest(s: String) {
    configuration.position2D match {
      case Some(p) => sendURC(List(SetNavigationDestination(s)))
        waiting4TargetReachedStateChange = true
        log.info("Got SetNavigation and sent to Rolland: " + s)
      case _ => log.info("Got SetNavigation, but waiting for position2D: " + s)
    }
    configuration.currentNavigationDestination = Some(s)
  }

  // do not change this timing; this is critical and synchronized with Wheelchairs
  context.setReceiveTimeout(50.millisecond)

  override def receive = {
    case ReceiveTimeout => obtainRollandInfo()
    case "stop" =>
      resetSocket()
      context.stop(self)
    case SetNavigationDestination(s) => setNaviDest(s)
    case VirtualID(n) =>
      configuration.virtualid = Some(n)
      informListenersAboutCurrentPosition()
    case CurrentOntology(o) if o == EmptyOntology =>
      configuration.reset()
      waiting4TargetReachedStateChange = false
      lastTargetReachedState = None
    case CurrentOntology(o) => {
      log.warning("RollandClient gots ontology")
      configuration.virtualid match {
        case Some(n) => {
          log.info("virtualid " + n)
          val ontpos = Var("mypos")
          log.info("testing " + atRole.roleAssert(Const(n), ontpos) + " = " + o.query(atRole.roleAssert(Const(n), ontpos)))
          o.query(atRole.roleAssert(Const(n), ontpos)) match {
            case s :: _ => s(ontpos) match {
              case Const(pos) if pos != "Hide" // TODO: This is not nice, we should have a Concept HidePosition and check that in the ontology
              => // we are not hidden, so let's see if we have a nextposition and it is already assigend
                log.warning(n + "Rolland pos " + pos)
                o.query(nextposRole.roleAssert(Const(n), Var("mynextpos"))) match {
                  case Nil => // no next pos, nothing to do
                    configuration.position match {
                      case Some(l) if l != pos && configuration.currentNavigationDestination.isEmpty =>
                        configuration.position = Some(l)
                        o.query(unassignedConcept.mkAssertion(Var("nowc"))) match {
                          case snowc :: _ =>
                            val nowc = snowc(Var("nowc")).asInstanceOf[Const].name
                            informListenersAboutCurrentPosition(withstatus = true, obsoletepos = Some((pos, nowc)))
                          case _ => log.warning("Cannot reset position_status for position " + pos + " because there is no Unassigned individual in the ontology")
                        }
                      case _ =>
                    }
                  case subst :: _ =>
                    val nextpos@(Const(np)) = subst(Var("mynextpos")).asInstanceOf[Const]
                    log.warning(n + " next pos " + nextpos)
                    o.query(posstatusRole.roleAssert(nextpos, Const(n))) match {
                      case Nil =>
                        log.warning(configuration.virtualid + " position not yet assigned")
                      // next pos not yet assigned, do nothing
                      case _ => // Next pos assigned to us, check if we are not already on the way
                        if (configuration.currentNavigationDestination == Some(np)) {
                          log.warning(configuration.virtualid + " already on the way")
                          // we are already on the way
                        } else {
                          // we are not on the way or at least not to this new position. Let's go...
                          log.warning(configuration.virtualid + " not on the way")
                          setNaviDest(np)
                        }
                    }
                }
              case Const(_) => // do nothing
              case Var(x) => // the ontology does not know where we are, lets inform it
                log.info(n + " has no virtual pos")
                configuration.currentNavigationDestination = None
                informListenersAboutCurrentPosition()
            }
            case Nil => // the ontology does not know where we are, lets inform it
              if (!waiting4TargetReachedStateChange) informListenersAboutCurrentPosition()
          }
        }
        case None => None
      }
    }
  }

  def informListenersAboutCurrentPosition(withstatus: Boolean = false, obsoletepos: Option[(String, String)] = None) {
    //println("listeners are "+listeners)
    configuration.position match {
      case Some(label) =>
        val update = if (withstatus)
          Update(
            List(RoleAssertion(Const(name), Const(label), atRole),
              posstatusRole.roleAssert(Const(label), Const(name))) ++
              (obsoletepos match {
                case Some((old, nowc)) => List(posstatusRole.roleAssert(Const(old), Const(nowc)))
                case None => Nil
              }), Nil)
        else Update(List(RoleAssertion(Const(name), Const(label), atRole)), List())
        listeners map (l => l ! update)
      case v =>
    }
  }
}
