package src.navi;


import java.awt.geom.Point2D;
import java.io.StringWriter;
import java.util.*;
import java.util.logging.Logger;

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 6/8/11
 * Time: 4:38 PM
 * This class represents a Map as provided by SimRobot. Edges and way points can be added in an arbitrary order, at the
 * cost of an additional check during insertion. This could be optimized for larger maps.
 */
public class WPMap {
    public static Logger logger = Logger.getLogger("Map");
    private int approxbezier;
    private Point2D upperleftcorner;
    private Point2D upperrightcorner;
    private Point2D downleftcorner;
    private Point2D upperleftcornerexternal;
    private Point2D upperrightcornerexternal;
    private Point2D downleftcornerexternal;
    private ArrayList<WayPoint> waypoints;

    public Map<WayPoint,List<WPEdge>> getEdges() {
        HashMap<WayPoint,List<WPEdge>> result = new HashMap<WayPoint,List<WPEdge>>();
        for(WayPoint wp:waypoints) {
            result.put(wp, wp.getEdges());
        }
        return result;
    }

    public int getApproxbezier() {
        return approxbezier;
    }


    public void setApproxbezier(int approxbezier) {
        this.approxbezier = approxbezier;
    }

    public Point2D getUpperleftcorner() {
        return upperleftcorner;
    }

    public void setUpperleftcorner(Point2D upperleftcorner) {
        this.upperleftcorner = upperleftcorner;
    }

    public Point2D getUpperrightcorner() {
        return upperrightcorner;
    }

    public void setUpperrightcorner(Point2D upperrightcorner) {
        this.upperrightcorner = upperrightcorner;
    }

    public Point2D getDownleftcorner() {
        return downleftcorner;
    }

    public void setDownleftcorner(Point2D downleftcorner) {
        this.downleftcorner = downleftcorner;
    }

    public Point2D getUpperleftcornerexternal() {
        return upperleftcornerexternal;
    }

    public void setUpperleftcornerexternal(Point2D upperleftcornerexternal) {
        this.upperleftcornerexternal = upperleftcornerexternal;
    }

    public Point2D getUpperrightcornerexternal() {
        return upperrightcornerexternal;
    }

    public void setUpperrightcornerexternal(Point2D upperrightcornerexternal) {
        this.upperrightcornerexternal = upperrightcornerexternal;
    }

    public Point2D getDownleftcornerexternal() {
        return downleftcornerexternal;
    }

    public void setDownleftcornerexternal(Point2D downleftcornerexternal) {
        this.downleftcornerexternal = downleftcornerexternal;
    }

    public WPMap() {
        waypoints = new ArrayList<WayPoint>();
    }

    public void addWayPoint(String id, int x, int y)
    {
        //logger.debug("Add wp "+id);
        WayPoint wp = findWP(id);
        if (wp==null) waypoints.add(new WayPoint(id,x,y));
        else {
            wp.setX(x);
            wp.setY(y);
        }
    }

    public WayPoint findWP(String name)
    {
        for (int i=0; i<waypoints.size(); i++)
            if (waypoints.get(i).getName().equals(name)) return waypoints.get(i);
        return null;
    }

    public void addEdge(String sourcewp, String targetwp, String type)
    {
        //logger.debug("Add edge " + sourcewp + " " + targetwp + " " + type);
        WayPoint wp = findWP(sourcewp);
        if (wp==null)
        {
            addWayPoint(sourcewp, 0, 0);
            wp=waypoints.get(waypoints.size()-1);
        }
        wp.addEdge(targetwp, type);
    }

    public String toString()
    {
        String result="";
        for (WayPoint wp : waypoints) result=result+" "+wp;
        return result;
    }

    public String toDot()
    {
        StringWriter sw = new StringWriter();
        sw.write("graph graphname {\n");
        for (WayPoint wp : waypoints)
        {
            for (WPEdge edge : wp.getEdges())
                sw.write(wp.getName()+" -- "+edge.getTarget()+"\n");
        }
        sw.write("}");
        return sw.toString();
    }

    public String toSVG()
    {
        StringWriter sw = new StringWriter();
        for (WayPoint wp : waypoints)
        {
            sw.write("<circle cx=\""+wp.getX()+"\" cy=\""+wp.getY()+"\" r=\"40\" stroke=\"black\" stroke-width=\"2\" fill=\"red\"/>\n");
            sw.write("<text x=\""+wp.getX()+"\" y=\""+wp.getY()+"\" style=\"font-size:60px\">"+wp.getName()+"</text>");
            for (WPEdge edge : wp.getEdges())
            {
                WayPoint target = findWP(edge.getTarget());
                sw.write( "<line x1=\""+wp.getX()+"\" y1=\""+wp.getY()+"\" x2=\""+target.getX()+"\" y2=\""+target.getY()+"\" style=\"stroke:rgb(255,0,0);stroke-width:2\"/>\n");
            }
        }
        return sw.toString();
    }


    class WPEdgeComparator implements Comparator<WPEdge>{
        public int compare( WPEdge a, WPEdge b )
        {
            if (findWP(a.getTarget()).getEdges().size() > findWP(b.getTarget()).getEdges().size()) return 1;
            else if (findWP(a.getTarget()).getEdges().size() == findWP(b.getTarget()).getEdges().size()) return 0;
            else return -1;
        }
    }


    /**
     * computes spanning tree that maximizes its nodes
     */
    public void computeSpanningTree()
    {
        Set<WayPoint> reachablenodes = new HashSet<WayPoint>();
        WayPoint currentwp = findWP("Start");
        ArrayList<WPEdge> edges = new ArrayList<WPEdge>();
        for (WayPoint wp : waypoints) edges.addAll(wp.getEdges());
        Comparator<WPEdge> comparator = new WPEdgeComparator();
        java.util.Collections.sort(edges, comparator);
        System.out.println("edges: "+edges);

    }
}


