package src.ship

import java.io.File

import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.io.{FileDocumentSource, IRIDocumentSource}
import org.semanticweb.owlapi.model._

import scala.collection.{JavaConversions, mutable}

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 25.04.2013
 * Time: 17:22
 */
object OWL2Ontology extends JavaLogging {

  /*
   Sample owl ontologies
   -  http://www.co-ode.org/ontologies/pizza/2005/10/18/pizza.owl
   -  http://www.scs.ryerson.ca/~bgajdero/msc_thesis/code/ontologies/weather-ont-t2.owl
   - file:///Users/autexier/Downloads/openAAL_1.1/OntologyCreator/OntologyFile/OpenAALOntology.owl
   - file:///Users/autexier/Repositories/autexier/lib/java/pellet-2.3.0/examples/data/people+pets.owl
   - file:///Users/autexier/Repositories/autexier/lib/java/pellet-2.3.0/examples/data/simple-galen.owl
   - file:///Users/autexier/Repositories/autexier/lib/java/pellet-2.3.0/examples/data/univ-bench.owl
   - file:///Users/autexier/Repositories/autexier/lib/java/pellet-2.3.0/examples/data/wine.owl
   - http://protege.cim3.net/file/pub/ontologies/generations/generations.owl

   Look also at http://www.co-ode.org/ontologies/


SubObjectPropertyOf(<http://www.co-ode.org/ontologies/pizza/2005/10/18/pizza.owl#hasTopping>
<http://www.co-ode.org/ontologies/pizza/2005/10/18/pizza.owl#hasIngredient>)
   */

  def fromIRIString(s: String): Option[Ontology] = {
    val iri = IRI.create(s)
    fromIRI(iri)
  }

  def fromIRI(ontiri: IRI): Option[Ontology] = {
    var result: Option[Ontology] = None
    try {
      val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
      var config = new OWLOntologyLoaderConfiguration()
      config = config.setMissingImportHandlingStrategy(MissingImportHandlingStrategy.SILENT)
      val owlont = manager.loadOntologyFromOntologyDocument(new IRIDocumentSource(ontiri), config)
      result = Some(convert2Ontology(owlont))
    }
    catch {
      case e: OWLOntologyCreationException => log.warning(e.toString)
    }
    result
  }

  def fromFile(file: File): Option[Ontology] = {

    var result: Option[Ontology] = None
    try {
      val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
      var config = new OWLOntologyLoaderConfiguration()
      config = config.setMissingImportHandlingStrategy(MissingImportHandlingStrategy.SILENT)
      val owlont = manager.loadOntologyFromOntologyDocument(new FileDocumentSource(file), config)
      result = Some(convert2Ontology(owlont))
    }
    catch {
      case e: OWLOntologyCreationException => log.warning(e.toString)
    }
    result
  }

  def convert2Ontology(owlont: OWLOntology): Ontology = {
    var tbox: mutable.Set[TBoxFormula] = mutable.Set.empty[TBoxFormula]
    var rbox: mutable.Set[RBoxFormula] = mutable.Set.empty[RBoxFormula]
    var abox: mutable.Set[ABoxFormula] = mutable.Set.empty[ABoxFormula]

    val owlaxioms = JavaConversions.asScalaSet(owlont.getAxioms)
    var usedaxioms = Set.empty[OWLAxiom]
    var annos: mutable.Map[OWLObject, OWLAnnotation] = mutable.Map.empty[OWLObject, OWLAnnotation]

    for (owlaxiom: OWLAxiom <- owlaxioms) {
      owlaxiom match {
        case ax: OWLAnnotationAssertionAxiom => annos = annos + (ax -> ax.getAnnotation)
        case _ =>
      }
      val tboxaxs = convertTBoxAxiom(owlaxiom)
      if (tboxaxs.nonEmpty) {
        tbox.++=(tboxaxs)
        usedaxioms = usedaxioms + owlaxiom
      }
      val rboxaxs = convertRBoxAxiom(owlaxiom)
      if (rboxaxs.nonEmpty) {
        rbox.++=(rboxaxs)
        usedaxioms = usedaxioms + owlaxiom
      }
      val aboxaxs = convertABoxAxiom(owlaxiom)
      if (aboxaxs.nonEmpty) {
        abox.++=(aboxaxs)
        usedaxioms = usedaxioms + owlaxiom
      }
    }

    aggregateRoleDeclarations(rbox)

    if (owlont.getAnnotations.size == 1) annos = annos + (owlont -> JavaConversions.asScalaSet(owlont.getAnnotations).head)

    log.debug("Ignored OWL Axioms:")
    for (owlax <- (owlaxioms -- usedaxioms).filterNot(p => p.isInstanceOf[OWLDeclarationAxiom] || p.isInstanceOf[OWLAnnotationAssertionAxiom])) {
      log.debug(" ignored " + owlax)
    }

    var o = Ontology(tbox.toList, rbox.toList, abox.toList)
    o.addAnnotations(annos)
    o
  }

  def aggregateRoleDeclarations(rbox: mutable.Set[RBoxFormula]) {
    val roledecls: mutable.Set[RoleDeclaration] = rbox.filter(phi => phi.isInstanceOf[RoleDeclaration]).asInstanceOf[mutable.Set[RoleDeclaration]]
    val duplicateroledecls = roledecls.iterator.scanLeft(Set[Role]())((set, a) => set + a.role) // incremental sets
      .zip(roledecls.iterator)
      .filter {
      case (set, a) => set contains a.role
    }
    while (duplicateroledecls.hasNext) {
      val role = duplicateroledecls.next()._2.role
      //println("Duplicate role decls "+role)
      val decls = rbox.filter {
        case RoleDeclaration(r, _, _) if r == role => true
        case _ => false
      }
      val domain = (decls.map {
        case RoleDeclaration(_, d, _) => d
        case _ => throw new Exception("aggregateRoleDeclarations1")
      } :\ (UniversalConcept: Concept)) {
        (c, d) => if (d == UniversalConcept) c else if (c == UniversalConcept) d else AndConcept(c, d)
      }
      val range = (decls.map {
        case RoleDeclaration(_, _, r) => r
        case _ => throw new Exception("aggregateRoleDeclarations2")
      } :\ (UniversalConcept: Concept)) {
        (c, d) => if (d == UniversalConcept) c else if (c == UniversalConcept) d else AndConcept(c, d)
      }
      rbox.--=(decls)
      rbox.+=(RoleDeclaration(role, domain, range))
    }
  }

  def convertTBoxAxiom(owl: OWLAxiom): Set[TBoxFormula] = {
    owl match {
      case owlax: OWLDisjointClassesAxiom =>
        (JavaConversions.asScalaSet(owlax.asPairwiseAxioms) :\ (Set.empty[TBoxFormula]: Set[TBoxFormula])) {
          (owleqax, set) => {
            val it = owleqax.getClassExpressions.iterator()
            val c1 = convertConcept(it.next())
            val c2 = convertConcept(it.next())
            set + ConceptEquality(NotConcept(UniversalConcept), AndConcept(c1, c2))
          }
        }
      case owlax: OWLDisjointUnionAxiom =>
        convertTBoxAxiom(owlax.getOWLEquivalentClassesAxiom) ++ convertTBoxAxiom(owlax.getOWLDisjointClassesAxiom)
      case owlax: OWLSubClassOfAxiom =>
        Set(ConceptInclusion(convertConcept(owlax.getSubClass), convertConcept(owlax.getSuperClass)))
      case owlax: OWLEquivalentClassesAxiom =>
        (JavaConversions.asScalaSet(owlax.asPairwiseAxioms) :\ (Set.empty[TBoxFormula]: Set[TBoxFormula])) {
          (owleqax, set) => {
            val it = owleqax.getClassExpressions.iterator()
            val c1 = convertConcept(it.next())
            val c2 = convertConcept(it.next())
            (c1, c2) match {
              case (ConceptConstant(_), _) => set + ConceptEquality(c1, c2)
              case (_, ConceptConstant(_)) => set + ConceptEquality(c2, c1)
              case _ => set + ConceptEquality(c1, c2)
            }
          }
        }
      case _ => Set()
    }
  }

  def convertRBoxAxiom(owl: OWLAxiom): Set[RBoxFormula] = {
    owl match {
      case owlax: OWLSubObjectPropertyOfAxiom =>
        val sub = convertRole(owlax.getSubProperty)
        val sup = convertRole(owlax.getSuperProperty)
        Set(RoleInclusion(List(sub), sup))
      case owlax: OWLSubPropertyChainOfAxiom =>
        val subs = (JavaConversions.asScalaBuffer(owlax.getPropertyChain) :\ (List.empty[Role]: List[Role])) {
          (owlrole, list) => convertRole(owlrole) :: list
        }
        val sup = convertRole(owlax.getSuperProperty)
        Set(RoleInclusion(subs, sup))
      case owlax: OWLObjectPropertyCharacteristicAxiom =>
        val role = convertRole(owlax.getProperty)
        owl match {
          case _: OWLReflexiveObjectPropertyAxiom => Set(RoleRef(role))
          case _: OWLIrreflexiveObjectPropertyAxiom => Set(RoleIrr(role))
          case _: OWLTransitiveObjectPropertyAxiom => Set(RoleTra(role))
          case _: OWLFunctionalObjectPropertyAxiom => Set(RoleFun(role))
          case _: OWLInverseFunctionalObjectPropertyAxiom => Set(RoleInvFun(role))
          case _: OWLSymmetricObjectPropertyAxiom => Set(RoleSym(role))
          case _: OWLAsymmetricObjectPropertyAxiom => Set(RoleAsym(role))
          case _ =>
            log.warning("Cannot translate OWLObjectPropertyCharacteristicAxiom " + owl)
            Set()
        }
      case owlax: OWLFunctionalDataPropertyAxiom =>
        val role = convertDataRole(owlax.getProperty)
        Set(RoleFun(role))
      case owlax: OWLObjectPropertyDomainAxiom =>
        val role = convertRole(owlax.getProperty)
        val concept = convertConcept(owlax.getDomain)
        Set(RoleDeclaration(role, concept, UniversalConcept))
      case owlax: OWLObjectPropertyRangeAxiom =>
        val role = convertRole(owlax.getProperty)
        val concept = convertConcept(owlax.getRange)
        Set(RoleDeclaration(role, UniversalConcept, concept))
      case owlax: OWLDataPropertyDomainAxiom =>
        val role = convertDataRole(owlax.getProperty)
        val concept = convertConcept(owlax.getDomain)
        Set(RoleDeclaration(role, concept, UniversalConcept))
      case owlax: OWLDataPropertyRangeAxiom =>
        val role = convertDataRole(owlax.getProperty)
        val concept = convertDataConcept(owlax.getRange)
        Set(RoleDeclaration(role, UniversalConcept, concept))
      case owlax: OWLInverseObjectPropertiesAxiom =>
        val r1 = convertRole(owlax.getFirstProperty)
        val r2 = convertRole(owlax.getSecondProperty)
        (r1, r2) match {
          case (NamedRole(_), _) => Set(RoleEquality(List(InverseRole(r2)), r1))
          case _ => Set(RoleEquality(List(InverseRole(r1)), r2))
        }
      case _ => Set()
    }
  }

  def getIDorName(i: OWLIndividual): Const = Const(
    if (i.isNamed) i.asOWLNamedIndividual().getIRI.getFragment
    else i.asOWLAnonymousIndividual().getID.getID)

  def convertABoxAxiom(owl: OWLAxiom): Set[ABoxFormula] = {
    owl match {
      case owlax: OWLClassAssertionAxiom =>
        val concept = convertConcept(owlax.getClassExpression)
        val ind = getIDorName(owlax.getIndividual)
        Set(concept.mkAssertion(ind))
      case owlax: OWLObjectPropertyAssertionAxiom =>
        val role = convertRole(owlax.getProperty)
        val src = getIDorName(owlax.getSubject)
        val trg = getIDorName(owlax.getObject)
        Set(role.roleAssert(src, trg))
      case owlax: OWLDataPropertyAssertionAxiom =>
        val role = convertDataRole(owlax.getProperty)
        val src = getIDorName(owlax.getSubject)
        val litObj = owlax.getObject
        val litType = litObj.getDatatype
        val lit = litObj.getLiteral
        val trg = Literal(if (litType.isBuiltIn && litType.getBuiltInDatatype.isNumeric) lit else "\"" + lit + "\"")
        Set(role.roleAssert(src, trg))
      case owlax: OWLSameIndividualAxiom =>
        (JavaConversions.asScalaSet(owlax.asPairwiseAxioms()) :\ Set[ABoxFormula]()) {
          (ax, set) => set + SameAsAssertion(convertIndividual(ax.getIndividualsAsList.get(0)),
            convertIndividual(ax.getIndividualsAsList.get(1)))
        }
      case owlax: OWLDifferentIndividualsAxiom =>
        (JavaConversions.asScalaSet(owlax.asPairwiseAxioms()) :\ Set[ABoxFormula]()) {
          (ax, set) => set + DifferentFromAssertion(convertIndividual(ax.getIndividualsAsList.get(0)),
            convertIndividual(ax.getIndividualsAsList.get(1)))
        }
      case _: OWLDisjointClassesAxiom => Set()
      case _ =>
        Set()
    }
  }

  def convertIndividual(ind: OWLIndividual): Const = {
    Const(ind.asOWLNamedIndividual().getIRI.getFragment)
  }

  def convertDataConcept(dr: OWLDataRange): Concept = {
    if (dr.isDatatype) {
      val n = dr.asOWLDatatype.getBuiltInDatatype.getShortName
      DataConcept(n, Nil, Nil)
    }
    else {
      log.warning("translation of OWL data concept not supported: " + dr)
      UniversalConcept
    }
  }

  def convertConcept(oc: OWLClassExpression): Concept = {
    oc match {
      case _ if oc.isOWLThing => UniversalConcept
      case _ if oc.isOWLNothing => NotConcept(UniversalConcept)
      case ono: OWLNamedObject => ConceptConstant(ono.getIRI.getFragment)
      case ooo: OWLObjectOneOf =>
        (JavaConversions.asScalaSet(ooo.getIndividuals) :\ (UniversalConcept: Concept)) {
          (owlind, c) => {
            val ind = Const(owlind.asOWLNamedIndividual().getIRI.getFragment)
            c match {
              case UniversalConcept => Nominal(ind)
              case _ => OrConcept(Nominal(ind), c)
            }
          }
        }
      case oio: OWLObjectIntersectionOf =>
        val subclasses = JavaConversions.asScalaSet(oio.getOperands)
        (subclasses :\ (UniversalConcept: Concept))((owlexp, c) => {
          c match {
            case UniversalConcept => convertConcept(owlexp)
            case _ => AndConcept(convertConcept(owlexp), c)
          }
        })
      case ouo: OWLObjectUnionOf =>
        val subclasses = JavaConversions.asScalaSet(ouo.getOperands)
        (subclasses :\ (UniversalConcept: Concept))((owlexp, c) => {
          c match {
            case UniversalConcept => convertConcept(owlexp)
            case _ => OrConcept(convertConcept(owlexp), c)
          }
        })
      case oco: OWLObjectComplementOf => convertConcept(oco.getOperand).Negate
      case oavf: OWLObjectAllValuesFrom =>
        val role = convertRole(oavf.getProperty)
        val concept = convertConcept(oavf.getFiller)
        ForallRole(role, concept)
      case osvf: OWLObjectSomeValuesFrom =>
        val role = convertRole(osvf.getProperty)
        val concept = convertConcept(osvf.getFiller)
        ExistsRole(role, concept)
      case ohv: OWLObjectHasValue =>
        val role = convertRole(ohv.getProperty)
        val ind = convertIndividual(ohv.getValue)
        ExistsRole(role, Nominal(ind))
      case ocr: OWLObjectCardinalityRestriction =>
        val card = ocr.getCardinality
        val role = convertRole(ocr.getProperty)
        val concept = convertConcept(ocr.getFiller)
        oc match {
          case _: OWLObjectMaxCardinality => MaxRestriction(role, card, concept)
          case _: OWLObjectMinCardinality => MinRestriction(role, card, concept)
          case _: OWLObjectExactCardinality => AndConcept(MinRestriction(role, card, concept), MaxRestriction(role, card, concept))
        }
      case dsvf: OWLDataSomeValuesFrom =>
        val role = convertDataRole(dsvf.getProperty)
        val concept = convertDataConcept(dsvf.getFiller)
        ExistsRole(role, concept)
      case davf: OWLDataAllValuesFrom =>
        val role = convertDataRole(davf.getProperty)
        val concept = convertDataConcept(davf.getFiller)
        ForallRole(role, concept)
      case dcr: OWLDataCardinalityRestriction =>
        val card = dcr.getCardinality
        val role = convertDataRole(dcr.getProperty)
        val concept = convertDataConcept(dcr.getFiller)
        oc match {
          case _: OWLDataMaxCardinality => MaxRestriction(role, card, concept)
          case _: OWLDataMinCardinality => MinRestriction(role, card, concept)
          case _: OWLDataExactCardinality => AndConcept(MinRestriction(role, card, concept), MaxRestriction(role, card, concept))
        }
      case _ =>
        log.warning("translation of OWL concept not supported (yet)? " + oc)
        UniversalConcept
    }
  }

  def convertDataRole(op: OWLDataPropertyExpression): Role = {
    op match {
      case dataProperty: OWLDataProperty =>
        val n = dataProperty.getIRI.getFragment
        DataRole(n)
      case _ =>
        log.warning("translation of OWL data role not supported (yet)?" + op)
        UniversalRole
    }
  }

  def convertRole(op: OWLObjectPropertyExpression): Role = {
    /*OWLRestriction[OWLClassExpression,OWLObjectPropertyExpression,OWLClassExpression]*/
    op match {
      case objectProperty: OWLObjectProperty =>
        NamedRole(objectProperty.getIRI.getFragment)
      case objectInverseOf: OWLObjectInverseOf =>
        val subrole = convertRole(objectInverseOf.getInverse)
        InverseRole(subrole)
      case _ =>
        log.warning("translation of OWL role not supported: " + op)
        UniversalRole
    }
  }
}
