package src.ship


import scala.util.parsing.combinator.JavaTokenParsers

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 7.09.2012
 * Time: 16:40
 */

class OntMorphism {
  var cmap = Map.empty[ConceptConstant, Concept]
  var rmap = Map.empty[NamedRole, Role]
  var imap = Map.empty[Const, Const]

  def compose(g: OntMorphism): OntMorphism = {
    val f = new OntMorphism
    f.cmap = (g.cmap mapValues apply) ++ ((cmap -- g.cmap.keys) mapValues g.apply)
    f.rmap = (g.rmap mapValues apply) ++ ((rmap -- g.rmap.keys) mapValues g.apply)
    f.imap = (g.imap mapValues applyToConst) ++ ((imap -- g.imap.keys) mapValues g.applyToConst)
    f
  }

  def apply(c: Concept): Concept = {
    c match {
      case ConceptVariable(_) => c
      case v@ConceptConstant(n) => cmap get v match {
        case Some(trg) => trg
        case None => c
      }
      case OrConcept(c1, c2) => OrConcept(apply(c1), apply(c2))
      case AndConcept(c1, c2) => AndConcept(apply(c1), apply(c2))
      case ForallRole(r, c1) => ForallRole(apply(r), apply(c1))
      case ExistsRole(r, c1) => ExistsRole(apply(r), apply(c1))
      case MaxRestriction(r, i, c1) => MaxRestriction(apply(r), i, apply(c1))
      case MinRestriction(r, i, c1) => MinRestriction(apply(r), i, apply(c1))
      case NotConcept(c1) => NotConcept(apply(c1))
      case At(v, c1) => At(apply(v), apply(c1))
      case Nominal(v) => Nominal(apply(v))
      case _ => c
    }
  }

  def applyToConst(v: Const): Const =
    v match {
      case c@Const(_) => imap get c match {
        case Some(x) => x
        case None => c
      }
  }

  def apply(v: VarOrConstant): VarOrConstant =
    v match {
      case x@Const(_) => applyToConst(x)
      case Var(x) => Const(x)
      case Literal(_) => v
  }

  def apply(r: Role): Role = r match {
      case v@NamedRole(n) => rmap get v match {
        case Some(trg) => trg
        case None => r
      }
      case AndRole(r1, r2) => AndRole(apply(r1), apply(r2))
      case OrRole(r1, r2) => OrRole(apply(r1), apply(r2))
      case InverseRole(r1) => InverseRole(apply(r1))
      case NotRole(r1) => NotRole(apply(r1))
      case NominalRole(v1, v2) => NominalRole(apply(Const(v1)).name, apply(Const(v2)).name)
      case _ => r
    }

  def apply(phi: TBoxFormula): TBoxFormula = phi match {
      case ConceptInclusion(sub, sup) => ConceptInclusion(apply(sub), apply(sup))
      case ConceptEquality(left, right) => ConceptEquality(apply(left), apply(right))
    }

  def apply(phi: RBoxFormula): RBoxFormula = phi match {
      case RoleDeclaration(r, dom, range) => RoleDeclaration(apply(r), apply(dom), apply(range))
      case RoleInclusion(rs, r) => RoleInclusion(rs map apply, apply(r))
      case RoleEquality(rs, r) => RoleEquality(rs map apply, apply(r))
      case RoleFun(r) => RoleFun(apply(r))
      case RoleTra(role) => RoleTra(apply(role))
      case RoleRef(role) => RoleRef(apply(role))
      case RoleIrr(role) => RoleIrr(apply(role))
      case RoleDis(r1, r2) => RoleDis(apply(r1), apply(r2))
      case RoleInvFun(role) => RoleInvFun(apply(role))
      case RoleAsym(role) => RoleAsym(apply(role))
      case RoleSym(role) => RoleSym(apply(role))
    }

  def apply(phi: ABoxFormula): ABoxFormula = phi match {
      case ConceptAssertion(n, c) => ConceptAssertion(apply(n), apply(c))
      case RoleAssertion(src, trg, r) => RoleAssertion(apply(src), apply(trg), apply(r))
      case DifferentFromAssertion(v1, v2) => DifferentFromAssertion(apply(v1), apply(v2))
      case SameAsAssertion(v1, v2) => SameAsAssertion(apply(v1), apply(v2))
      case _ => println("Morphism not implemented for " + phi)
        phi
    }

  def apply(o: Ontology): Ontology =
    Ontology(o.tbox map apply, o.rbox map apply, o.abox map apply)
}

case class ConceptMorph(src: ConceptConstant, trg: Concept) extends OntMorphism {
  cmap = Map[ConceptConstant, Concept] {
    src -> trg
  }
}

case class RoleMorph(src: NamedRole, trg: Role) extends OntMorphism {
  rmap = Map[NamedRole, Role] {
    src -> trg
  }
}

case class IndMorph(src: Const, trg: Const) extends OntMorphism {
  imap = Map[Const, Const] {
    src -> trg
  }
}

case object IdMorph extends OntMorphism

trait OntMorphismParser extends JavaTokenParsers
with TBoxFormulaParser
with RBoxFormulaParser
with ABoxFormulaParser {
  protected override val whiteSpace = """(\s|//.*|(?m)/\*(\*(?!/)|[^*])*\*/)+""".r

  def ontmorphparser: Parser[OntMorphism] = conceptmorphparser ~ rolemorphparser ~ indmorphparser ^^ {
    case cmorph ~ rmorph ~ imorph => imorph.compose(rmorph.compose(cmorph))
  }

  def conceptmorphparser: Parser[OntMorphism] = {
    opt("concepts" ~ repsep(string ~ ("|->" | "as") ~ concept, ",")) ^^ {
      case Some(_ ~ cmaps) =>
        var res: OntMorphism = IdMorph
        for (cmap <- cmaps) {
          //println("cmap ="+cmap)
          cmap match {
            case src ~ _ ~ trg => res = res.compose(ConceptMorph(ConceptConstant(src), trg))
            case _ =>
          }
        }
        res
      case None => IdMorph
    }
  }

  def rolemorphparser: Parser[OntMorphism] = {
    opt("roles" ~ repsep(string ~ ("|->" | "as") ~ role, ",")) ^^ {
      case Some(_ ~ rmaps) =>
        var res: OntMorphism = IdMorph
        for (rmap <- rmaps) {
          rmap match {
            case (src: String) ~ _ ~ trg => res = res.compose(RoleMorph(NamedRole(src), trg))
            case _ =>
          }
        }
        res
      case None => IdMorph
    }
  }

  def indmorphparser: Parser[OntMorphism] = {
    opt("individuals" ~ repsep(string ~ ("|->" | "as") ~ string, ",")) ^^ {
      case Some(_ ~ imaps) =>
        var res: OntMorphism = IdMorph

        for (imap <- imaps) {
          imap match {
            case (src: String) ~ _ ~ trg => res = res.compose(IndMorph(Const(src), Const(trg)))
            case _ =>
          }
        }
        res
      case None => IdMorph
    }
  }
}
