package src.ship

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:03
 */

trait ConceptOrRole extends ObjectLevelVariable {
  def getConcepts: Set[ConceptConstant] = Set.empty

  def getRoles: Set[NamedRole] = Set.empty

  def getIndividuals: Set[String] = Set.empty

  def isAtomic = false

  def isLiteral = false
}

sealed trait Role extends ConceptOrRole with GroundableObject[Role] {

  def roleAssert(src: VarOrConstant, trg: VarOrConstant): RoleAssertion = RoleAssertion(src, trg, this)

  def isNegationOf(r: Role): Boolean = {
    this match {
      case UniversalRole => r match {
        case NotRole(UniversalRole) => true
        case _ => false
      }
      case NamedRole(n) => {
        r match {
          case NotRole(NamedRole(n1)) => n == n1
          case _ => false
        }
      }
      case VariableRole(n) => {
        r match {
          case NotRole(VariableRole(n1)) => n == n1
          case _ => false
        }
      }
      case InverseRole(r2) => {
        r match {
          case NotRole(r1) => this == r1
          case _ => false
        }
      }
      case NominalRole(s, t) => {
        r match {
          case NotRole(NominalRole(s1, t1)) => s == s1 && t == t1
          case _ => false
        }
      }
      case NotRole(r1) => r == r1
      case _ => false
    }
  }

  def Negate: Role = NotRole(this)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = this
}

case object UniversalRole extends Role {
  override def toString: String = "\u0394\u00D7\u0394"

  override def isAtomic = true

  override def isLiteral = true
}

case class NamedRole(name: String) extends Role {
  var namespace = Builtin.namespace

  override def toString: String = name

  override def isAtomic = true

  override def isLiteral = true

  override def getRoles: Set[NamedRole] = Set(this)

  def setNamespace(s: String): NamedRole = {
    namespace = s
    this
  }
}

case class VariableRole(name: String) extends Role {
  override def toString: String = name.toUpperCase

  override def freeVars = Set(this)

  override def isAtomic = true

  override def isLiteral = true

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = NamedRole(name)
}

case class DataRole(name: String) extends Role {
  override def toString: String = name

  override def isAtomic = true

  override def isLiteral = true
}

case class NotRole(sub: Role) extends Role {
  override def toString: String = "~" + (if (sub.isLiteral) sub else "(" + sub + ")")

  override def freeVars = sub.freeVars

  // not Role must be encoded

  override def isLiteral = sub.isAtomic

  override def Negate = sub

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = NotRole(sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = sub.getConcepts

  override def getRoles: Set[NamedRole] = sub.getRoles

  override def getIndividuals: Set[String] = sub.getIndividuals
}

case class InverseRole(sub: Role) extends Role {
  override def toString: String = "inv(" + sub + ")"

  override def freeVars = sub.freeVars

  override def isAtomic = sub.isAtomic

  override def isLiteral = sub.isAtomic

  override def Negate = InverseRole(sub.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = InverseRole(sub.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = sub.getConcepts

  override def getRoles: Set[NamedRole] = sub.getRoles

  override def getIndividuals: Set[String] = sub.getIndividuals
}

case class AndRole(left: Role, right: Role) extends Role {
  override def toString: String = "(" + left + "\u2293" + right + ")"

  override def freeVars = left.freeVars ++ right.freeVars

  override def Negate = OrRole(left.Negate, right.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = AndRole(left.groundBy(exceptions), right.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = left.getConcepts ++ right.getConcepts

  override def getRoles: Set[NamedRole] = left.getRoles ++ right.getRoles

  override def getIndividuals: Set[String] = left.getIndividuals ++ right.getIndividuals
}

case class OrRole(left: Role, right: Role) extends Role {
  override def toString: String = "(" + left + "\u2294" + right + ")"

  override def freeVars = left.freeVars ++ right.freeVars

  override def Negate = AndRole(left.Negate, right.Negate)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = OrRole(left.groundBy(exceptions), right.groundBy(exceptions))

  override def getConcepts: Set[ConceptConstant] = left.getConcepts ++ right.getConcepts

  override def getRoles: Set[NamedRole] = left.getRoles ++ right.getRoles

  override def getIndividuals: Set[String] = left.getIndividuals ++ right.getIndividuals
}

case class NominalRole(src: String, trg: String) extends Role {
  override def toString: String = "{(" + src + ", " + trg + ")}"

  override def freeVars = Set.empty[ObjectLevelVariable]

  override def isAtomic = true

  override def isLiteral = true

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = NominalRole(src, trg)

  override def getIndividuals: Set[String] = Set(src, trg)
}
