package src.ship.external

import collection.mutable
import scala.Some
import src.BAALL.{RollandConfiguration, DrivingDestinations, ConflictMap}
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 12.09.12
 * Time: 14:00
 */
object PathCoordination extends JavaLogging {

  private var currentConflictMap = {
    DrivingDestinations(new RollandConfiguration().drivingDestinations).conflicts
  }

  def setCurrentConflictMap(m: ConflictMap) {
    currentConflictMap = m
  }

  def getCurrentConflictMap: ConflictMap = currentConflictMap

  /**
   * computes a sequence of (0/1,0/1) where x_i=0 means that robot 0 must stop and x_i=1 means that robot is allowed
   * to move. we introduce intermediate waypoints between two points that indicate that the go was already given
   * @param route1 route of the first robot, including the current position
   * @param route2 route of the second robot, including the current position
   * @return a sequence of moves
   */
  def coordinateRoutes(route1: Array[String], route2: Array[String]): List[(Int, Int)] = {
    log.info("coordinate routes for " + route1.mkString(",") + "\nand " + route2.mkString(",") +
      "\nconflict map is " + currentConflictMap)
    val matrix = Array.ofDim[Int](route1.length * 2, route2.length * 2)
    for (i <- 0 to route1.length * 2 - 1)
      for (j <- 0 to route2.length * 2 - 1) {
        val seg1start = route1(i / 2)
        val seg1next = if (i / 2 >= route1.length - 1) route1(i / 2) else route1(i / 2 + 1)
        val seg1go = i % 2 != 0
        val seg2start = route2(j / 2)
        val seg2next = if (j / 2 + 1 >= route2.length - 1) route2(j / 2) else route2(j / 2 + 1)
        val seg2go = j % 2 != 0

        matrix(i)(j) =
          if (
            segmentoverlaps(seg1start, seg1next, seg1go, seg2start, seg2next, seg2go)
              ||
              !seg1go && positionOverlapsWithSegment(seg1start, seg2start, seg2next, seg2go)
              ||
              !seg2go && positionOverlapsWithSegment(seg2start, seg1start, seg1next, seg1go)
              ||
              seg1go && positionOverlapsWithSegment(seg1next, seg2start, seg2next, seg2go)
              ||
              seg2go && positionOverlapsWithSegment(seg2next, seg1start, seg1next, seg1go)
          ) 1
          else 0
      }
    closeMatrix(matrix, route1.length * 2, route2.length * 2)
    //extractDistanceMap(matrix, route1.length*2, route2.length*2)
    log.info("matrix:\n" +
      matrix.map(_.mkString(" ")).mkString("\n") +
      "\nr1: " + route1.length * 2 + " r2:" + route2.length * 2)
    extractPathRec(matrix, 0, 0, route1.length * 2, route2.length * 2, Nil) match {
      case Nil => Nil
      case v@(solution :: _) => {
        log.info("#solutions = " + v.length)
        solution
      }
    }

  }

  /**
   * computes the convex hull using a sweep line algorithm and modifies the matrix
   * @param matrix the original matrix
   * @param l1 rows
   * @param l2 columns
   */
  def closeMatrix(matrix: Array[Array[Int]], l1: Int, l2: Int) {
    // boundary closure
    var bl = 0
    for (j <- 0 to l2 - 1) if (matrix(l1 - 1)(j) > 0) bl = j
    if (bl > 0) for (j <- 0 to bl) matrix(l1 - 1)(j) = 3

    bl = 0
    for (i <- 0 to l1 - 1) if (matrix(i)(l2 - 1) > 0) bl = i
    if (bl > 0) for (i <- 0 to bl) matrix(i)(l2 - 1) = 3


    var i: Int = l1 - 2
    while (i >= 0) {
      var j = l2 - 2
      while (j >= 0) {
        if (matrix(i)(j + 1) > 0 && matrix(i + 1)(j) > 0 && matrix(i)(j) == 0) matrix(i)(j) = 2
        if (i > 0 && matrix(i)(j) == 0 && matrix(i - 1)(j) > 0 && matrix(i + 1)(j) > 0) matrix(i)(j) = 2
        if (j > 0 && matrix(i)(j) == 0 && matrix(i)(j - 1) > 0 && matrix(i)(j + 1) > 0) matrix(i)(j) = 2
        j = j - 1
      }
      i = i - 1
    }

  }

  /**
   *
   * @param matrix path matrix
   */
  def extractPath(matrix: Array[Array[Int]], r1length: Int, r2length: Int): List[(Int, Int)] = {
    // every Scala collection has a mkString() method that takes a string delimiter and joins the values in the collection
    log.info("matrix:\n" +
      matrix.map(_.mkString(" ")).mkString("\n"))
    var posi = 0
    var posj = 0
    log.info("r1: " + r1length + " r2:" + r2length)
    var solution = List[(Int, Int)]()
    while (posi < r1length - 2 || posj < r2length - 2) {
      log.info("i " + posi + " j " + posj + " sol: " + solution + " r1:" + r1length + " r2:" + r2length)
      if (posj == r2length - 2) {
        if (matrix(posi + 1)(posj) == 0 && matrix(posi + 2)(posj) == 0) {
          posi = posi + 2
          solution = solution ++ List((1, 0))
        }
        else {
          log.info("solution at " + posi + " " + posj + " not possible case 1")
          return List()
        }
      }
      else if (posi == r1length - 2) {
        if (matrix(posi)(posj + 1) == 0 && matrix(posi)(posj + 2) == 0) {
          posj = posj + 2
          solution = solution ++ List((0, 1))
        }
        else {
          log.info("solution at " + posi + " " + posj + " not possible case 2")
          return List()
        }
      }
      else if (matrix(posi)(posj) == 0 && matrix(posi + 1)(posj) == 0 &&
        matrix(posi)(posj + 1) == 0 && matrix(posi + 1)(posj + 1) == 0 &&
        matrix(posi + 2)(posj + 2) == 0) {
        solution = solution ++ List((1, 1))
        posi = posi + 2
        posj = posj + 2
      }
      else if (matrix(posi)(posj + 1) == 0 && matrix(posi)(posj + 2) == 0) {
        solution = solution ++ List((0, 1))
        posj = posj + 2
      }
      else if (matrix(posi + 1)(posj) == 0 && matrix(posi + 2)(posj) == 0) {
        solution = solution ++ List((1, 0))
        posi = posi + 2
      }
      else {
        log.info("solution at " + posi + " " + posj + " not possible case 3")
        return List()
      }
    }
    solution
  }

  def extractPathRec(matrix: Array[Array[Int]], posi: Int, posj: Int, r1length: Int, r2length: Int, solution: List[(Int, Int)]): List[List[(Int, Int)]] = {
    // every Scala collection has a mkString() method that takes a string delimiter and joins the values in the collection
    if (posi < r1length - 2 || posj < r2length - 2) {
      log.info("i " + posi + " j " + posj + " sol: " + solution + " r1:" + r1length + " r2:" + r2length)
      if (posj == r2length - 2) {
        if (matrix(posi + 1)(posj) == 0 && matrix(posi + 2)(posj) == 0) {
          extractPathRec(matrix, posi + 2, posj, r1length, r2length, solution ++ List((1, 0)))
        }
        else {
          log.warning("solution at " + posi + " " + posj + " not possible case 1")
          List()
        }
      }
      else if (posi == r1length - 2) {
        if (matrix(posi)(posj + 1) == 0 && matrix(posi)(posj + 2) == 0) {

          extractPathRec(matrix, posi, posj + 2, r1length, r2length, solution ++ List((0, 1)))
        }
        else {
          log.warning("solution at " + posi + " " + posj + " not possible case 2")
          List()
        }
      }
      else if (matrix(posi)(posj) == 0 && matrix(posi + 1)(posj) == 0 &&
        matrix(posi)(posj + 1) == 0 && matrix(posi + 1)(posj + 1) == 0 &&
        matrix(posi + 2)(posj + 2) == 0) {
        extractPathRec(matrix,
          posi + 2,
          posj + 2,
          r1length, r2length, solution ++ List((1, 1)))
      }
      else {
        val l =
          (if (matrix(posi)(posj + 1) == 0 && matrix(posi)(posj + 2) == 0) {
            extractPathRec(matrix, posi,
              posj + 2, r1length, r2length, solution ++ List((0, 1)))
          } else Nil) ++
            (if (matrix(posi + 1)(posj) == 0 && matrix(posi + 2)(posj) == 0) {
              extractPathRec(matrix, posi + 2, posj, r1length, r2length, solution ++ List((1, 0)))
            } else Nil)
        l match {
          case Nil => log.warning("solution at " + posi + " " + posj + " not possible case 3")
          case _ =>
        }
        l
      }
    }
    else List(solution)
  }

  /**
   *
   * @param matrix path matrix
   */
  def extractDistanceMap(matrix: Array[Array[Int]], r1length: Int, r2length: Int): Array[Array[Int]] = {
    val distmatrix = Array.ofDim[Int](r1length, r2length)
    log.info(matrix.map(_.mkString(" ")).mkString("\n"))
    distmatrix(0)(0) = 1
    for (i <- 1 to r1length - 1) if (matrix(i)(0) == 0 && distmatrix(i - 1)(0) > 0) distmatrix(i)(0) = distmatrix(i - 1)(0) + 1
    else distmatrix(i)(0) = -1

    for (j <- 1 to r2length - 1) if (matrix(0)(j) == 0 && distmatrix(0)(j - 1) > 0) distmatrix(0)(j) = distmatrix(0)(j - 1) + 1
    else distmatrix(0)(j) = -1

    for (i <- 1 to r1length - 1)
      for (j <- 1 to r2length - 1) {
        log.info("i: " + i + " j: " + j)
        if (matrix(i)(j) > 0) distmatrix(i)(j) = -1
        else
        if (distmatrix(i - 1)(j - 1) < 0 && distmatrix(i - 1)(j) < 0 && distmatrix(i)(j - 1) < 0) distmatrix(i)(j) = -1
        else {
          if (distmatrix(i - 1)(j - 1) > 0) distmatrix(i)(j) = distmatrix(i - 1)(j - 1) else distmatrix(i)(j) = 100000
          if (distmatrix(i - 1)(j) > 0 && distmatrix(i - 1)(j) < distmatrix(i)(j)) distmatrix(i)(j) = distmatrix(i - 1)(j)
          if (distmatrix(i)(j - 1) > 0 && distmatrix(i)(j - 1) < distmatrix(i)(j)) distmatrix(i)(j) = distmatrix(i)(j - 1)
          distmatrix(i)(j) = distmatrix(i)(j) + 1
        }
      }
    log.info(distmatrix.map(_.mkString(" ")).mkString("\n"))
    distmatrix
  }

  /**
   * we define overlapping sements that originate at waypoint i and include the next field in the direction of the path
   * @param seg1start current wp, even means that it is an actual wp, odd means it is an imaginary wp
   * @param seg1next next wp or current wp if wc is idle or has reached its goal
   * @param seg2start current wp
   * @param seg2next next wp or current wp if wc is idle or has reached its goal
   * @return true if there might be a collision
   */
  def segmentoverlaps(seg1start: String, seg1next: String, seg1go: Boolean,
                      seg2start: String, seg2next: String, seg2go: Boolean): Boolean = {
    if (seg1go) {
      // seg1go given
      if (seg2go) {
        //seg2go given
        seg1start == seg2start || seg2next == seg1start || seg1next == seg2start || seg1next == seg2next
      }
      else {
        //seg2go not given
        seg1start == seg2start || seg1next == seg2start
      }
    }
    else {
      // seg1go not given
      if (seg2go) {
        //seg2go given
        seg1start == seg2start || seg2next == seg1start
      }
      else {
        //seg2go not given
        seg1start == seg2start
      }
    }
  }

  def positionOverlapsWithSegment(seg1: String, seg2start: String, seg2next: String, seg2go: Boolean): Boolean = {
    if (seg2go) {
      val x = currentConflictMap.conflicts(seg1, seg2start, seg2next)
      if (x) log.info("Position " + seg1 + " conflicts with drive from " + seg2start + " to " + seg2next)
      x
    } else {
      val x = currentConflictMap.conflicts(seg1, seg2start)
      if (x) log.info("Position " + seg1 + " conflicts with position " + seg2start)
      x
    }

  }

  def computeNewSchedule(ont: Ontology): Option[UnconditionalEffect] = {

    // 1. Get all wheelchairs and routes
    val substs = OWLQuery.execQuery(
      "SELECT ?wc ?r ?wcpos WHERE {  Type(?wc,ship:WheelChair), " +
        "Type(?r,ship:Route), " +
        "PropertyValue(?wc,ship:wheelchair_route,?r)," +
        "PropertyValue(?wc,ship:at,?wcpos) }",
      ont.getManager, ont.getReasoner,
      "ship", Builtin.namespace)
    //println("substs is "+substs.mkString(";;"))
    var wheelchair2route = Map.empty[String, List[String]]
    var wheelchairs = List.empty[String]
    for (x <- substs) {
      val wc = x.apply(Var("wc")).name
      val startpos = x.apply(Var("wcpos")).name
      val wcroute = startpos :: Ontology2Scala.RoleEntryAsList(ont, NamedRole("route_next"), NamedRole("route_rest"), x.apply(Var("r")).name)
      log.info("wcroute for " + wc + " is " + wcroute)
      wheelchair2route = wheelchair2route.updated(wc, wcroute)
      wheelchairs = wc :: wheelchairs
    }
    //println("Wheelchairs = "+wheelchairs)
    val schedules = OWLQuery.execQuery("SELECT ?s WHERE {  Type(?s,ship:WCSchedule)}", ont.getManager, ont.getReasoner, "ship", Builtin.namespace)
    val directive = OWLQuery.execQuery("SELECT ?s WHERE {  Type(?s,ship:WCDirective)}", ont.getManager, ont.getReasoner, "ship", Builtin.namespace)
    val directives = OWLQuery.execQuery("SELECT ?s WHERE {  Type(?s,ship:WCDirectives)}", ont.getManager, ont.getReasoner, "ship", Builtin.namespace)
    // 2. Delete currentschedule

    // 3. Compute new schedule; if exists, insert schedule, otherwise insert NewSimpleSchedulingPossible
    val pairlist2array = (l: List[(String, String)]) => {
      l match {
        case Nil => Array[String]()
        case _ => (l.head._1 :: (l map (p => p._2))).toArray[String]
      }
    }

    val route1 =
      if (wheelchairs.size >= 1) wheelchair2route.get(wheelchairs.apply(0)).getOrElse(Nil).toArray[String]
      else new Array[String](0)
    val route2 =
      if (wheelchairs.size >= 2) wheelchair2route.get(wheelchairs.apply(1)).getOrElse(Nil).toArray[String]
      else new Array[String](0)


    val newschedule =
      PathCoordination.coordinateRoutes(route1, route2)
    log.info("Computed schedule for " +(wheelchairs.apply(0), wheelchairs.apply(1)) + "\n" + newschedule)
    val newabox = schedule2Ontology(ont, (wheelchairs.apply(0), wheelchairs.apply(1)), newschedule)
    Some(UnconditionalEffect(newabox,
      (schedules map ((s: Substitution) => s.apply(Var("s")))) ++
        (directives map ((s: Substitution) => s.apply(Var("s"))))
        ++ (directive map ((s: Substitution) => s.apply(Var("s"))))))
  }

  private def schedule2Ontology(ont: Ontology, wheelchairs: (String, String), schedule: List[(Int, Int)]): List[ABoxFormula] = {
    val newabox = new mutable.MutableList[ABoxFormula]

    wheelchairs match {
      case (wc1, wc2) =>

        val wcschedules: List[String] = schedule map (directive => directive match {
          case (dirwc1, dirwc2) =>
            val (wc1dirbox, dirwc1ind) = Scala2Ontology.pair2Abox((wc1, if (dirwc1 == 0) "stop" else "go"),
              "WCDirective", "wcdirective_wc", "wcdirective_directive", ont)
            val (wc2dirbox, dirwc2ind) = Scala2Ontology.pair2Abox((wc2, if (dirwc2 == 0) "stop" else "go"),
              "WCDirective", "wcdirective_wc", "wcdirective_directive", ont)
            val (wcdirectivesabox, wcdirectivesind) =
              Scala2Ontology.list2Abox(List[String](dirwc1ind, dirwc2ind), "", "rest_wcdirectives", "next_wcdirective",
                "NoWCDirectives", "WCDirectives", ont)
            newabox.++=(wc1dirbox)
            newabox.++=(wc2dirbox)
            newabox.++=(wcdirectivesabox)
            wcdirectivesind
        })
        val (schedulebox, scheduleind) = Scala2Ontology.list2Abox(wcschedules, "", "rest_schedule", "current_directives", "WCEmptySchedule", "WCSchedule", ont)
        newabox.++=(schedulebox)
        //newabox.+=(ConceptAssertion(Const(scheduleind),ConceptConstant("WCSchedule")))
        log.info("Generated schedule " + scheduleind + "\n" + newabox.toList)
        newabox.toList
    }
  }
}
